from logging import Logger
from urllib.request import Request

from flask.wrappers import Response

from pbench.server import OperationCode, PbenchServerConfig
from pbench.server.api.resources import (
    ApiAuthorizationType,
    ApiBase,
    ApiContext,
    ApiMethod,
    ApiParams,
    ApiSchema,
    Parameter,
    ParamType,
    Schema,
)
from pbench.server.auth.auth import Auth
from pbench.server.database.models.datasets import Dataset, DatasetError


class DatasetOwner(ApiBase):
    """
    API class to retrieve inventory files from a dataset
    """

    def __init__(self, config: PbenchServerConfig, logger: Logger):
        super().__init__(
            config,
            logger,
            ApiSchema(
                ApiMethod.PUT,
                OperationCode.READ,
                uri_schema=Schema(
                    Parameter("dataset", ParamType.DATASET, required=True),
                ),
                authorization=ApiAuthorizationType.DATASET,
            ),
        )

    def _put(
        self, params: ApiParams, request: Request, context: ApiContext
    ) -> Response:
        """
        This function returns the contents of the requested file as a byte stream.

        Args:
            params: includes the uri parameters, which provide the dataset and target.
            request: Original incoming Request object
            context: API context dictionary

        Raises:
            DatasetError, reporting "NOT_FOUND"


        PUT /api/v1/datasets/change_ownership/{dataset}
        """

        dataset = params.uri["dataset"]
        try:
            dataset = Dataset.attach(resource_id=dataset)
        except DatasetError as e:
            self.logger.error(
                "Unable to find dataset with resource ID {!r}: {}",
                dataset,
                str(e),
            )

        authorized_id = Auth.token_auth.current_user().id

        dataset.update(owner_id=authorized_id)
        self.logger.info("Ownership of the dataset has been modified to {}", dataset)
